(*
 * SNU 4190.310 Programming Languages (Spring 2011)
 *
 * M0: definition of syntax, runner, and pretty printer
 *)

type mexp = Num of int
          | Var of id
          | Fn of id * mexp
          | App of mexp * mexp
		  | Rec of id * id * mexp
          | Ifz of mexp * mexp * mexp
		  | Add of mexp * mexp
		  | Sub of mexp * mexp
          | Pair of mexp * mexp    (* (e, e) *)
          | Fst of mexp            (*   e.1  *)
          | Snd of mexp            (*   e.2  *)
and id = string

exception RuntimeError of string
exception TypeError of string

module M_Runner = struct
	type loc = int
	type value = N of int
			   | P of value * value
			   | Closure of closure
	and closure = fexpr * env
	and fexpr = F of id * mexp
			  | R of id * id * mexp
	and env = id -> value
	type memory = int * (loc -> value)

	let (@+) f (x, v) = (fun y -> if y=x then v else f y)
	let store (l, m) p = (l, m @+ p)
	let fetch (_, m) l = m l
	let malloc (l, m) = (l, (l+1, m))

    let error msg = raise (RuntimeError msg)
    let getClosure = function (Closure c) -> c | _ -> error "not a function"
    let getNum = function (N n) -> n | _ -> error "not a number value"
    let getPair = function (P (a,b)) -> (a, b) | _ -> error "not a pair"

    let rec eval env mem exp = match exp
        with Num n -> (N n, mem)
         | Var x -> (env x, mem)
         | Fn (x, e) -> (Closure (F (x, e), env), mem)
		 | Rec (f, x, e) -> (Closure (R (f, x, e), env), mem)
         | App (e1, e2) ->
            let (v1, m') = eval env mem e1 in
            let (c, env') = getClosure v1 in
            let (v2, m'') = eval env m' e2 in
                (match c with F (x, e) -> eval (env' @+ (x,v2)) m'' e
                        | R (f, x, e) -> eval ((env' @+ (x,v2)) @+ (f,Closure(c,env'))) m'' e)
         | Ifz (e1, e2, e3) ->
            let (v1, m') = eval env mem e1 in
				eval env m' (if (getNum v1 = 0) then e2 else e3)
         | Pair(e1, e2) ->
            let (v1, m') = eval env mem e1 in
            let (v2, m'') = eval env m' e2 in
            (P (v1, v2), m'')
         | Fst e ->
            let (v, m') = eval env mem e in
            let (v1, v2) = getPair v in
            (v1, m')
         | Snd e ->
            let (v, m') = eval env mem e in
            let (v1, v2) = getPair v in
            (v2, m')
		 | Add (e1, e2) ->
			let (v1, m) = eval env mem e1 in
			let (v2, m') = eval env m e2 in
			(N (getNum v1 + getNum v2), m')
		 | Sub (e1, e2) ->
			let (v1, m) = eval env mem e1 in
			let (v2, m') = eval env m e2 in
			(N (getNum v1 - getNum v2), m')

	let emptyEnv = (fun x -> error ("unknown id: " ^ x))
	let emptyMem = (0, fun l -> error ("unknown loc: " ^ string_of_int l))
	
	let evalm exp =
		let (v,_) = eval emptyEnv emptyMem exp in
		getNum v

end


module M_Printer = struct
    let ps = print_string
    let nl = print_newline
    let indent i =
        let rec it = function 0 -> ()
        		   | n -> ps " "; it (n-1)
        in  nl (); it i
        
    let rec pp n =
        function 
         | Num i -> print_int i
         | Var s -> ps s
         | Fn (x, e) -> ps ("fn "^x^" => "); (
                 match e with
                   Fn _ -> pp (n+1) e
				 | Rec _ -> pp (n+1) e
                 | _ -> indent (n+1); pp (n+1) e
             )
         | App (e, e') -> pp n e; ps " "; pp n e'
         | Ifz (e1, e2, e3)-> ps "ifzero "; pp n e1; ps " then ";
                          indent (n+1); pp (n+1) e2;
                          indent (n); ps "else";
                          indent (n+1); pp (n+1) e3
         | Pair (e1, e2) -> ps "("; pp n e1; ps ", "; pp n e2; ps ")"
         | Fst e -> pp n e; ps ".1"
         | Snd e -> pp n e; ps ".2"
         | Add (e1, e2) -> ps "("; pp n e1; ps " + "; pp n e2; ps ")"
         | Sub (e1, e2) -> ps "("; pp n e1; ps " - "; pp n e2; ps ")"
		 | Rec (f, x, e) -> ps ("rec "^f^" "^x^" => "); (
				match e with
                   Fn _ -> pp (n+1) e
				 | Rec _ -> pp (n+1) e
                 | _ -> indent (n+1); pp (n+1) e
             )

    let print = pp 0
end
